<?php
/**
 * Here you can configure which template will be used for which check command.
 *
 * Normally, the graphs use a template that matches the name of your
 * check command, or the template "default" if nothing else exists.
 *
 * This page simplifies setting up better templates for those checks that don't
 * already have good defaults.
 *
 * Better templates might compact multiple datasources into one single,
 * easy-to-understand graph, format domain-specific information better, or
 * attempt to predict when thresholds will be breached.
 *
 * For creating your own templates, please see http://docs.pnp4nagios.org/pnp-0.6/tpl
 *
 * When you have configured your Graph template, the graph will be
 * visible in the Graphs UI within seconds. This means that you do not need to
 * apply your changes when dealing with Graph templates, they are changed as
 * soon as you save them.
 */

class graph_template_nacoma_Model extends pnp_nacoma_Model
{
	protected static $base = '/opt/monitor/op5/pnp/'; /**< Base path to PNP */
	protected static $dist = 'templates.dist/'; /**< Directory of distributed templates */
	protected static $user = 'templates/'; /**< Directory of user-provided templates */

	public $properties = array(
		'var_index' => array(
			'check' => 'select',
			'graph_template' => 'select',
		),
		'required' => array(
			'check',
			'graph_template',
		),
		'label' => array(
			'check' => 'Check command',
			'graph_template' => 'Graph template',
		),
		'special' => array(
			'check' => 'command',
			'graph_template' => array(),
		),
		'default' => array(
			'graph_template' => 'default',
		),
		'topic' => 'check',
	);

	const TYPE = 'graph_template';

	/**
	 * Return the mayi resource for the object
	 * @see object_nacoma_Model::mayi_resource()
	 */
	public function mayi_resource() {
		return "monitor.monitoring.graph_templates";
	}

	function __construct($id = false)
	{
		$this->type = static::TYPE;
		$this->get_properties();
		if (!$id)
			return;

		if (!is_numeric($id)) {
			$this->get_object_by_name($id);
		} else {
			$this->id = $id;
			$this->get_object();
		}
	}

	function get_properties()
	{
		# locate any templates.
		# first, locate any distributed (in the .dist folder).
		# then, because we only backup files in the non-dist-folder, we can't require
		# templates to be there. So also look for non-symlink-files in the template dir.
		#
		# you can't both select a template, and have a user-provided check with the
		# same name, but that makes no sense, so that's not my problem.

		$tmpls = opendir(self::$base . self::$dist);
		while ($template = readdir($tmpls)) {
			if (strpos($template, '.php') === false)
				continue;
			# if user has written their own custom template, ignore that
			if (!file_exists(self::$base . self::$user . $template) || !is_link(self::$base . self::$user . $template))
				$this->properties['special']['graph_template'][self::$dist . $template] = self::$dist . $template;
		}
		closedir($tmpls);

		$tmpls = opendir(self::$base . self::$user);
		while ($template = readdir($tmpls)) {
			if (strpos($template, '.php') === false)
				continue;
			if (!is_link(self::$base . self::$user . $template))
				$this->properties['special']['graph_template'][self::$user . $template] = self::$dist . $template;
		}
		closedir($tmpls);
		asort($this->properties['special']['graph_template']);
		return $this->properties;
	}

	static public function object_exists($id)
	{
		return command_nacoma_Model::object_exists($id);
	}

	function get_object_by_name($name)
	{
		$id = command_nacoma_Model::get_id_for($name);
		if (!$id) {
			return false;
		}
		if (file_exists(self::get_pnp_template_name($name))) {
			$this->id = $id;
			return $this->get_object();
		}
		return false;
	}

	/**
	 * Given a check name, returns the path to where a user should stick their
	 * template for it.
	 */
	protected static function get_pnp_template_name($check)
	{
		$check = static::objname_to_pnp($check);
		return self::$base . self::$user . $check . '.php';
	}

	function get_object()
	{
		$cmd = new command_nacoma_Model($this->id);
		$name = $cmd->get_object_name();
		if (!file_exists(self::get_pnp_template_name($name)))
			return false;
		if (!($link = realpath(self::get_pnp_template_name($name))))
			return false;
		if ($link === self::get_pnp_template_name($name))
			return false;
		if (strpos($link, self::$base) !== 0)
			return false;
		$this->obj['check'] = $this->id;
		$this->obj['graph_template'] = substr($link, strlen(self::$base));
		return $this->obj;
	}

	function delete_object()
	{
		if (!$this->id || $this->id == 'new') {
			$this->delete_message = "What is this magic? Why can't I find an object to delete?";
			return false;
		}
		if (is_numeric($this->id)) {
			$cmd = new command_nacoma_Model($this->id);
			$name = $cmd->get_object_name();
		} else {
			$name = basename($this->id);
		}
		if (!file_exists(self::get_pnp_template_name($name))) {
			$this->delete_message = 'Graph template does not exist';
			return false;
		}
		if (!is_link(self::get_pnp_template_name($name))) {
			$this->delete_message = 'Graph template is not a symbolic link, refusing to delete';
			return false;
		}

		if (!unlink(self::get_pnp_template_name($name))) {
			$this->delete_message = 'Unknown error deleting template';
			return false;
		}
		$this->delete_message = "Graph template $name deleted";
		return true;
	}

	function save_object($user=true)
	{
		if (!is_array($this->obj) || !$this->obj['check']) {
			return $this->validation_error('check', "Couldn't find anything to save");
		}
		$obj = $this->denormalize();
		if (!isset($obj['graph_template'])) {
			return $this->validation_error('graph_template', 'No template to use configured');
		} else if (!is_file(self::$base . $obj['graph_template'])) {
			return $this->validation_error('graph_template', 'Template does not exist');
		}
		# Clean up old, if it makes sense
		if ($this->obj['check'] !== $this->id && $this->id && $this->id !== 'new') {
			if (!$this->delete_object())
				return $this->validation_error('permission', "Couldn't delete previous file name {$this->id}, refusing to write new copy with other name: " . $this->delete_message);
		}
		# Woha, new one already in place? Well, it could be we're re-saving this one
		if (file_exists(self::get_pnp_template_name($obj['check']))) {
			if ($this->obj['check'] !== $this->id)
				return $this->validation_error('check', "Custom template for {$obj['check']} already exists");
			else if (!$this->delete_object())
				return $this->validation_error('permission', "Couldn't delete previous file, can't re-write it then: " . $this->delete_message);
		}
		if (!symlink(self::$base . $obj['graph_template'], self::get_pnp_template_name($obj['check'])))
			return $this->validation_error('check', 'Unknown error writing to '.self::get_pnp_template_name($obj['check']));

		$this->id = $this->obj['check'];
		return true;
	}

	public static function get_objects_for_user($write = false)
	{
		$res = array();
		foreach (command_nacoma_Model::get_objects_for_user() as $id => $cmd) {
			if (file_exists(self::get_pnp_template_name($cmd)) && is_link(self::get_pnp_template_name($cmd))) {
				$thecmd = basename($cmd, '.php');
				$res[$thecmd] = $thecmd;
			}
		}
		return $res;
	}

	function get_object_name()
	{
		if ($this->id && $this->id !== 'new') {
			$cmd = new command_nacoma_Model($this->id);
			$name = $cmd->get_object_name();
			$this->topic = $name;
			return $this->topic;
		}
		return 'New Custom Graph Template';
	}

	/**
	 * Make sure the controller is used instead of the generic delete.php
	 *
	 * @return array
	 */
	function get_cmd_bar_links() {
		$links = array();
		$out = parent::get_cmd_bar_links();
		if (isset($out['delete']))
			$links['delete'] = $out['delete'];
		return $links;
	}
}
