<?php

/**
 * The specific exception thrown when validation fails
 */
class ManagementPackManagementValidationException extends Exception {}

/**
 * Misc functionality to make working with management packs easier
 * Name mostly chosen because it's ridiculous
 */
class management_pack_management_nacoma_Model
{
	/**
	 * Given an object, return normalized data, free from external references
	 */
	private function clean_object($obj)
	{
		$obj->collapse_object();
		$data = $obj->denormalize();
		unset($data['id']);
		unset($data['register']);
		foreach ($obj->properties['special'] as $key => $type) {
			if ($obj->type == 'service' && ($key == 'hostgroup_name' || $key == 'check_command'))
				continue;
			if (!is_array($type))
				unset($data[$key]);
		}
		return $data;
	}

	/**
	 * Given data representing a "full" management pack (with commands etc), returns:
	 *    1 if the pack is already completely imported, and we can ignore it.
	 *    0 if the pack is importable, ie not imported at all, or only "simple" fields differs changed.
	 *    array of error messages if the pack appears modified and we must ask the user to be careful.
	 * This is really just the same code over and over, except with minor changes in variable names
	 * because everything is a special case.
	 */
	function is_imported($new_pack)
	{
		$errors = array();
		$res = 1;
		/*
		 * An obvious definition of "tricky state" is to have
		 * modified check commands. Those are easy to check,
		 * so check those first.
		 */
		foreach ($new_pack['commands'] as $command) {
			$old_cmd = new command_nacoma_Model($command['command_name']);
			if (!$old_cmd->id)
				$res = 0;
			else if ($old_cmd->obj['command_line'] != $command['command_line'])
				$errors[] = "The command '{$command['command_name']}' exists, but with different command line.<br/>\nNew: '{$command['command_line']}'<br/>\nOld: '{$old_cmd->obj['command_line']}'";
		}
		/*
		 * Great, then we can ignore checkcommands for the remainder of the method.
		 */

		$old_pack = new management_pack_nacoma_Model($new_pack['management_pack']['management_pack_name']);
		$old_hg = new hostgroup_nacoma_Model($new_pack['hostgroup']['hostgroup_name']);
		/*
		 * If the hostgroup exists, but not a management pack "owning" it,
		 * that's a sign of the user having Oppinions. So that's a minus one issue.
		 */
		if ($old_hg->id && !$old_pack->id)
			$errors[] = "There is already a hostgroup '{$new_pack['hostgroup']['hostgroup_name']}', but not a management pack '{$new_pack['management_pack']['management_pack_name']}' connected to it.";
		/*
		 * There are neither a hostgroup (therefore no services), no mgmt pack,
		 * and above we checked that no commands are modified. Let's rock!
		 */
		if (!$old_hg->id && !$old_pack->id)
			$res = 0;
		/*
		 * Let's check if the saved hg is the same. It shouldn't change.
		 *
		 * It's possible that this is stupid - it marks a changed hg note
		 * as trixy. We might lift these restrictions later?
		 */
		if ($old_hg->id) {
			$old_data = $old_hg->denormalize();
			foreach ($new_pack['hostgroup'] as $key => $val) {
				if (isset($old_data[$key]) && $val != $old_data[$key])
					$errors[] = "There is a hostgroup '{$old_data['hostgroup_name']}', yet its '$key' property is different from the one we're importing: new value '$val', old value '{$old_data[$key]}'.";
				else if (!isset($old_data[$key]))
					$res = 0;
			}
		}
		/*
		 * Mgmtpack vars are unimportant - everything but custom vars and hostgroup
		 * may change.
		 */
		if ($old_pack->id) {
			try {
				$old_data = $old_pack->denormalize();
			} catch (NacomaRelationException $ex) {
				$old_data = array();
			}
			foreach ($new_pack['management_pack'] as $key => $val) {
				if (isset($old_data[$key]) && $val != $old_data[$key]) {
					if ($key != 'hostgroup' || $key[0] == '_')
						$res = 0;
					else
						$errors[] = "The management pack '{$new_pack['management_pack']['management_pack_name']}' already exists, but it has a different value of $key from the one we're importing: new value '$val', old value '{$old_data[$key]}'.";
				}
				else if (!isset($old_data[$key])) {
					// New, I guess, custom var? It is by itself fine, it's up
					// to the commands to make this -1
					$res = 0;
				}
			}
		}
		/*
		 * Services. There might be new services, which is fine.
		 * There might be changed services, which is scary.
		 */
		foreach ($new_pack['services'] as $service) {
			$old_svc = new service_nacoma_Model();
			$old_svc->set_master_object_type('hostgroup');
			$old_svc->get_object_by_name($service['hostgroup_name'].';'.$service['service_description']);
			if (!$old_svc->id) {
				$res = 0;
				continue;
			}
			$old_data = $old_svc->denormalize();
			foreach ($service as $key => $val) {
				if (isset($old_data[$key]) && $val != $old_data[$key])
					$errors[] = "There is already a service '{$service['hostgroup_name']};{$service['service_description']}', but its $key property is different from what we want to import: new value '$val', old value '{$old_data[$key]}'.";
				else if (!isset($old_data[$key]))
					$res = 0;
			}
		}

		if ($errors)
			return $errors;
		return $res;
	}

	private function cleanup_failure($objs)
	{
		foreach ($objs as $obj) {
			if ($obj->id)
				revert_users_changes($obj->id);
		}
	}

	/**
	 * Given data representing a "full" mgmt pack (with commands etc), tries to import
	 * @param $pack_data The data to import
	 * @param $force If true, will import even if the user has modified the pack
	 * @return the new object id if the pack was imported, false if we bailed, an object containing validation errors in exceptional cases
	 */
	function import($pack_data, $force=false)
	{
		if (!$force && is_array($this->is_imported($pack_data)))
			return false;

		$objs = array();

		foreach ($pack_data['commands'] as $command) {
			$cmd = new command_nacoma_Model($command['command_name']);
			$cmd->get_default_object_sanely();
			$cmd->normalize($command);
			if (!$cmd->save_object()) {
				$this->cleanup_failure($objs);
				return $cmd;
			}
			$objs[] = $cmd;
		}
		$hg = new hostgroup_nacoma_Model($pack_data['hostgroup']['hostgroup_name']);
		$hg->get_default_object_sanely();
		$hg->normalize($pack_data['hostgroup']);
		if (!$hg->save_object()) {
			$this->cleanup_failure($objs);
			return $hg;
		}
		$objs[] = $hg;
		foreach ($pack_data['services'] as $service) {
			$svc = new service_nacoma_Model();
			$svc->set_master_object_type('hostgroup');
			$svc->get_object_by_name($service['hostgroup_name'].';'.$service['service_description']);
			$svc->get_default_object_sanely();
			$svc->normalize($service, false);
			if (!$svc->save_object()) {
				$this->cleanup_failure($objs);
				return $svc;
			}
			$objs[] = $svc;
		}
		$mgmt = new management_pack_nacoma_Model($pack_data['management_pack']['management_pack_name']);
		$mgmt->get_default_object(); // not using sanely: nothing interesting can ever be persisted, as it's just descriptions
		$mgmt->normalize($pack_data['management_pack'], false);
		if (!$mgmt->save_object()) {
			$this->cleanup_failure($objs);
			return $mgmt;
		}
		$objs[] = $mgmt;

		return $mgmt->id;
	}

	/**
	 * Given the name of a management pack, returns an array representing the "expanded" pack
	 * @returns an array of data on success, an error string on failure
	 */
	function export($pack_name, $description)
	{
		$mgmt_pack = new management_pack_nacoma_Model($_GET['management_pack_name']);
		if (!$mgmt_pack->id) {
			return "Couldn't find management pack";
		}
		$hg = new hostgroup_nacoma_Model($mgmt_pack->obj['hostgroup']);
		if (!$hg->id) {
			return "Couldn't find hostgroup";
		}
		$mgmt_pack->obj['version'] = isset($mgmt_pack->obj['version'])?$mgmt_pack->obj['version']+1:1;
		$full_svcs = $hg->find_slave_objects('service');
		$commands = array();
		$svcs = array();
		foreach ($full_svcs as $svc_id) {
			$svc = new service_nacoma_Model($svc_id);
			$svcs[] = $this->clean_object($svc);
			$command = new command_nacoma_Model($svc->obj['check_command']);
			// only export a single copy of every command
			$commands[$svc->obj['check_command']] = $this->clean_object($command);
		}
		$objs = array(
			'management_pack' => $this->clean_object($mgmt_pack),
			'hostgroup' => $this->clean_object($hg),
			'services' => $svcs,
			'commands' => array_values($commands)
		);
		$pack = array(
			'management_pack' => $objs,
			'changes' => array(
				array(
					'version' => $mgmt_pack->obj['version'],
					'description' => $description,
				),
			),
			'vendor' => '',
		);
		return $pack;
	}

	/**
	 * @throws Exception with reason
	 */
	static function validate_input($data) {
		if(!isset($data['management_pack']) || !is_array($data['management_pack']) ||
			!isset($data['management_pack']['management_pack']) || !is_array($data['management_pack']['management_pack']) ||
			!isset($data['management_pack']['management_pack']['management_pack_name']) ||
			!is_string($data['management_pack']['management_pack']['management_pack_name'])) {
			throw new ManagementPackManagementValidationException("Missing 'management_pack'");
		}
		if(!isset($data['management_pack']['commands']) ||
			!is_array($data['management_pack']['commands'])) {
			throw new ManagementPackManagementValidationException("Missing 'commands'");
		}
		if(!isset($data['management_pack']['services']) || !is_array($data['management_pack']['services'])) {
			throw new ManagementPackManagementValidationException("Missing 'management_pack'");
		}
		if(!isset($data['management_pack']['hostgroup']) || !is_array($data['management_pack']['hostgroup'])) {
			throw new ManagementPackManagementValidationException("Missing 'hostgroup'");
		}
		if(!isset($data['changes']) || !is_array($data['changes'])) {
			throw new ManagementPackManagementValidationException("Missing 'changes'");
		}
		if(!isset($data['vendor']) || !is_string($data['vendor'])) {
			throw new ManagementPackManagementValidationException("Missing 'vendor'");
		}
	}
}
