<?php

require_once('op5/log.php');

/**
 * Exception that is thrown for normalization errors
 */
class NacomaRelationException extends Exception {}

/**
 * Factory for specialized Nacoma errors that are easy to catch
 */
class NacomaExceptions {

	/**
	 * @param $obj_name string As user-friendly a name you can find for this object
	 * @param $key string The name of the attribute where the error occured
	 * @param $val string The value to the other object that couldn't be resolved
	 * @return NacomaRelationException
	 */
	static function invalid_key($obj_name, $key, $val) {
		return new NacomaRelationException("'$key' is not a valid attribute for '$obj_name'. You tried to set it to '$val'");
	}

	/**
	 * @param $obj_name string As user-friendly a name you can find for this object
	 * @param $key string The name of the attribute where the error occured
	 * @param $val string The value to the other object that couldn't be resolved
	 * @return NacomaRelationException
	 */
	static function invalid_value($obj_name, $key, $val) {
		return new NacomaRelationException("The attribute '$key', on object '$obj_name', could not be set to '$val', since '$val' was not found");
	}

	/**
	 * @param $obj_name string As user-friendly a name you can find for this object
	 * @param $key string The name of the attribute where the error occured
	 * @param $val string The value to the other object that couldn't be resolved
	 * @param $expected_type string The value to the other object that couldn't be resolved
	 * @return NacomaRelationException
	 */
	static function wrong_type($obj_name, $key, $val, $expected_type) {
		return new NacomaRelationException("The attribute '$key', on object '$obj_name', should be of the type ".help_nacoma_Model::translate_internal_property_type_to_external($expected_type));
	}
}

/**
 * generic object functions. Often replaced by object-specific
 * versions with the same name.
 *
 * This class is intended to be the one all the object-specific
 * classes extend, and as such it will hold generic functions
 * for pretty much everything.
 *
 * This is focused on nagios objects, but can be used for other objects as well.
 * For examples, see pnp_nacoma_Model + the graph_* subclasses
 */
abstract class object_nacoma_Model {
	/** The id of the current object */
	public $id = false;
	/** True if this object is a template, false otherwise */
	public $is_template = false;
	/** The object data in a key => value array, could be regarded as $values */
	public $obj = false;
	/** Presentation name for this object */
	public $topic = false;
	/**
	 * This object's properties (required vars, allowed values etc).
	 *
	 * The number of keys here is larger than the number of sticks up asses at
	 * a Republican Party convention, and it's absolutely central to nacoma that
	 * everything agrees on how it works. We usually ensure this invariant
	 * holds by never letting more than one person per year work on nacoma, so
	 * that person messes up consistently.
	 *
	 * The most important key is `var_index'. Var index holds an array of
	 * [field_name => form_value_type]. That is, the value must correspond to a
	 * form widget. Every member field of an object must exist here.
	 *
	 * There's the `default' key, that holds [field_name => default_value], for
	 * all field that has a default.
	 *
	 * There's also the `visible' key, that holds [field_name] for fields that
	 * should be rendered by default in "compact" mode. This mode only exists
	 * for non-template hosts, services and contacts, and it may thus be
	 * omitted.
	 *
	 * There's the `special' key. This should be set for "complicated" fields.
	 * Non-complicated fields are strings, numbers and similar. Complicated are
	 * checkboxes, selects, mselects, and similar. The field name here either
	 * maps to a string which is the name of another object type, or to an
	 * array of options.
	 *
	 * Among neat-but-optional keys, there's `topic'. It should only contain a
	 * string which has the name of the field which best describes this object,
	 * if any. Do not confuse it with $this->topic, which contains the actual
	 * name of the object, which probably used the `topic' key here to discover
	 * it.
	 *
	 * There's also `required', which absotively does not mean what the name
	 * suggests, because where would you get your kicks then, huh? A "required"
	 * variable, when nacoma talks about it, is a variable that must be
	 * resolvable, rather than one that must be provided - if it's set in a
	 * template, you do not need to provide it. If there's a hardcoded
	 * notification_interval in nagios that nacoma knows about, it can be
	 * "required" even though you're never required to provide it - True Story!
	 * This should probably change at some point.
	 *
	 * Another one is `widget'. This should contain [field_name => widget_name]
	 * for fields where you want to override the rendering widget, without
	 * replacing the validation widget. For example, you can use it to render a
	 * field as a hidden, yet validate it as a select relation.
	 *
	 * And the final one is `label'. This can contain
	 * [field_name => Pretty Name] to use a more user-friendly name in web
	 * forms, compared to the default, the field_name.
	 */
	public $properties = false;
	/** The master object id */
	public $master_obj_id = false;
	/** Type of the master object */
	public $master_obj_type = false;
	/** [type => key] If there are several possible master object types for this object, list them here */
	public $master_obj_candidates = false;
	/** Name of variable that holds the master object id */
	public $master_obj_identifier = false;
	/** Database table to use for this object */
	public $db_table = false;
	/** Custom variables present in this object */
	public $custom_vars = array(); # custom vars
	/** A list of object types that might depend on this object type, or false*/
	public $slave_types = false;
	/** If empty, consider validation as a) not run; or b) validated and safe */
	public $validation_error_list = array();
	/** Message when deleting or banishing objects */
	public $delete_message = false;
	/** The type name of this object - must be overridden */
	const TYPE = false;

	static private $images = false;

	abstract public function mayi_resource();

	/**
	 * constructor
	 */
	public function __construct($id = false) {
		$this->get_properties();
		$this->type = static::TYPE;
		if ($this->db_table === false)
			$this->db_table = $this->type;
		if (!$this->master_obj_candidates) {
			$this->master_obj_candidates =
				array($this->master_obj_type => $this->master_obj_identifier);
		}
		if ($id === false)
			return;

		if (is_numeric($id) && intval($id) == $id && $id > 0) {
			$this->id = intval($id);
			$this->get_object();
		} else {
			$this->get_object_by_name($id);
		}
	}

	/**
	 * Fetch images / icons, for icon selectors
	 */
	protected function get_images() {
		if (!self::$images) {
			self::$images = array();
			$img_list = glob("/opt/monitor/op5/ninja/application/media/images/logos/*");
			foreach ($img_list as $img) {
				self::$images[basename($img)] = basename($img);
			}
		}
		return self::$images;
	}


	/**	returns an object as identified by its nagios-name */
	public function get_object_by_name($name)
	{
		if (!$this->properties['topic'])
			return false;


		$query = 'SELECT id FROM ' . $this->db_table .
			" WHERE {$this->properties['topic']} = " . nacoma_db::quote_string($name);
		$result = nacoma_db::exec_query($query);

		if (!($row = $result->fetch()))
		{
			return false;
		}
		$this->id = $row[0];
		unset($result);
		return $this->get_object();
	}

	/**
	 * This method does not get defaults for your objects - of course not.
	 * It *throws away* your object, and *replaces* it with an empty "default object".
	 */
	public function get_default_object()
	{
		$this->obj = array();
		$this->custom_vars = array();

		$this->get_default_object_sanely();
	}

	/**
	 * This fills up your object with the defaults, which means it basically
	 * does an expand, as an expand against "no object" basically should give
	 * you a default object, but of course that's not how the world really
	 * works, because nacoma's defaults are totally different from nagios', and
	 * the notation for defaults in nacoma is "hilarious", in much the same way
	 * as shoving a spork into your own leg is.
	 *
	 * So it's an expand, with properties[default],
	 * and with every single line special cased because properties[default]
	 * expresses "default value" and "should be visible by default" in the same
	 * damn array.
	 */
	public function get_default_object_sanely($seen = array())
	{
		if (empty($this->properties['default']))
			return;
		if (in_array($this->id, $seen, true))
			return;
		$seen[] = $this->id;

		switch ($this->db_table) {
		 // the ones with templates
		 case 'service':
		 case 'host':
		 case 'contact':
			$tmpl_nacoma_Model = $this->db_table . '_template_nacoma_Model';
			$template = false;
			if (isset($this->obj['template'])) {
				$template = new $tmpl_nacoma_Model($this->obj['template']);
			} else if (isset($this->properties['default']['template'])) {
				$template = new $tmpl_nacoma_Model($this->properties['default']['template']);
			}
			if (!$template || !$template->id)
				return;

			$template->get_default_object_sanely($seen);

			if (!isset($this->obj['template']))
				$this->obj['template'] = $template->id;

			foreach ($template->obj as $k => $v) {
				switch ($k) {
				 case 'name':
				 case 'id':
				 case 'template':
					break;
				 default:
					if (!isset($this->obj[$k]) || $this->obj[$k] === '')
						$this->obj[$k] = $v;
				}
			}
			break;
		}
		$defaults = array();
		foreach ($this->properties['default'] as $key => $default) {
			if (is_array($this->obj) && isset($this->obj[$key]) && $this->obj[$key] !== '')
				continue;
			$defaults[$key] = $default;
		}

		$this->normalize($defaults, false);
	}

	/**
	 * Fetches the properties for the current object type.
	 */
	public function get_properties()
	{
		return $this->properties;
	}

	/**
	 * Propagates one or more variables to one or more objects
	 *
	 * @param $prop_vars The variables to propagate (k => v array)
	 * @param $obj_list Array of object-id's to propagate to
	 * @param $prop_format The type of value propagation to use for mselect variables
	 * @return True on success. Array of object validation errors otherwise
	 */
	public function propagate_variables($prop_vars, $obj_list, $prop_format='replace')
	{
		$errors = array();
		$class_name = $this->type . '_nacoma_Model';

		if (empty($prop_vars) || empty($obj_list))
			return false;
		foreach ($obj_list as $id) {
			$pobj = new $class_name($id);
			$pobj->expand_object();
			$props = $pobj->get_properties();
			$vidx = $props['var_index'];
			foreach ($prop_vars as $k => $v) {
				if (isset($vidx[$k]) && $vidx[$k] === 'mselect') {
					switch ($prop_format) {
					 case 'append':
						$dn = $pobj->denormalize();
						$res = $dn[$k];
						foreach ($v as $mk => $mv)
							$res[] = $mv;
						$v = $res;
						break;
					 case 'subtract':
						$dn = $pobj->denormalize();
						$v = array_diff($dn[$k], $v);
						break;
					 case 'replace':
					 default:
						break;
					}
				}
				try {
					$pobj->normalize(array($k => $v));
				} catch (NacomaRelationException $ex) {
					$errors[$pobj->get_object_name()] = array($k => $ex->getMessage());
				}
			}

			if ($pobj->save_object() !== true) {
				$errors[$pobj->get_object_name()] = $pobj->validation_error_list;
			}
		}

		return empty($errors) ? true : $errors;
	}

	/**
	 * Returns a presentation-friendly name of the object in question
	 */
	public function get_object_name()
	{
		if (!empty($this->topic))
			return $this->topic;

		if (!empty($this->obj)) {
			if ($this->id && !isset($this->obj['id']))
				$this->obj['id'] = $this->id;
			$this->topic = get_object_name_by_array($this->type, $this->obj);
		}
		else {
			$this->topic = get_object_name_by_id($this->type, $this->id);
		}
		return $this->topic;
	}

	/**
	 * Create a copy of an object. A copy is like a clone, except it has no
	 * identifying name, so it needs to be modified before being saved
	 * Generally useful for master objects.
	 */
	public function copy_object()
	{
		$nobj = $this->clone_object();
		$nobj->topic = get_object_name_by_array($nobj->type, array('id' => 'new'));
		if (isset($nobj->properties['topic']))
			$nobj->obj[$nobj->properties['topic']] = '';
		return $nobj;
	}

	/**
	 * Create a clone of an object. A clone can be saved straight away.
	 * Generally useful for slave objects.
	 * See $this->copy_object()
	 */
	public function clone_object()
	{
		$class = $this->type . '_nacoma_Model';
		$nobj = new $class();

		$ary = get_object_vars($this);
		foreach ($ary as $k => $v)
		  $nobj->$k = $v;

		$nobj->clone_id =  $ary['id'];
		if(isset($nobj->obj['command_name']))
			$nobj->clone_name = $nobj->obj['command_name'];
		$nobj->id = 'new';
		unset($nobj->obj['id']);

		return $nobj;
	}

	/**
	 * Check if an object of the specified type and id exists
	 *
	 * @param $obj_type string; Object type to check for
	 * @param $id integer; Id of the object to find
	 */
	static public function object_exists($id)
	{
		$class = static::TYPE . '_nacoma_Model';
		$obj = new $class($id);
		return !!$obj->id;
	}

	/**
	 * Helper method that returns an identifier, given a
	 * presentation name. This verifies permissions.
	 * @param $name string A user-presentable name (i.e. received from ->get_object_name())
	 * @returns An id, or false if none exists/was allowed. Note that id is object-specific - usually an int, but could be a string.
	 */
	static public function get_id_for($name)
	{
		if (!static::TYPE) {
			throw new Exception("get_id_for called on no type. This is a bug");
		}
		// This is quite inefficient, as it loads in
		// the *whole* object, just to look at one column and throw
		// all the others away again.
		// An optimized version could be provided only for nagios objects
		// which does the whole thing as a SELECT, given that we make db_table
		// static. The implementation is left as an exercise for the reader.
		$obj = new static;
		$obj->get_object_by_name($name);
		if ($obj->id && static::can_user_read($obj->id))
			return $obj->id;
		else
			return false;
	}

	/**
	 * Validates an object identifier (host_name, timeperiod_name, ...)
	 * for the most common cases. Note that not all objects have
	 * clearly defined identifiers.
	 */
	public function validate_object_identifier()
	{
		if (!isset($this->obj[$this->properties['topic']]) || !$this->obj[$this->properties['topic']]) {
			return $this->validation_error($this->properties['topic'], "{$this->properties['topic']} can not be empty");
		}

		$matches = false;
		// p{L} = letter, sounds reasonable, right? There are p{Lm}, modifier letters, which could be confusing, but meh.
		// p{N} = number, which doesn't sound too bad, does it?
		// p{S} = symbol, like currency and stuff - and p{Sm}, which is somewhat scary, but not overly so.
		// p{P} = punctuation. Somewhat scary, so exclude the ones banned by nagios.
		// p{M} = marks. Now we're getting fancy. But I did write a test for bengal, and I don't want to remove it.
		// p{Zs} = whitespace separator - but not paragraph or line separator.
		// p{Co} = private use. Seriously, it's absolutely vital we allow Klingon.
		$grouplist = '\p{L}\p{N}\p{S}\p{Zs}\p{Co}\p{P}\p{M}';
		if (preg_match_all("/[`~!$%^&*|'\"<>?,()=]/", $this->obj[$this->properties['topic']], $matches) || !preg_match('/^['.$grouplist.']+$/u', $this->obj[$this->properties['topic']])) {
			if (!$matches)
					preg_match_all('/[^'.$grouplist.']/u', $this->obj[$this->properties['topic']], $matches);
			// we don't want to comma separate this output, since comma itself is disallowed
			$str = "Illegal characters in {$this->properties['topic']}: ".implode(" ", $matches[0]);
			$this->validation_error($this->properties['topic'], $str);
		}
	}

	/**
	 * Always returns false for convenience, as it allows
	 * expressions like this:
	 *   return validation_error('everything', 'totally bogus stuff');
	 */
	public function validation_error($what, $why)
	{
		$this->validation_error_list[$what] = $why;
		return false;
	}

	/**
	 * returns the number of validation errors
	 */
	public function validation_error_count()
	{
		if (empty($this->validation_error_list))
			return 0;

		return count($this->validation_error_list);
	}

	/**
	 * Validates an object prior to saving it to database
	 *
	 * This function ensures that the object being validated:
	 *  - has all the required variables
	 *  - doesn't rely on an object that doesn't exist
	 *  - has valid input for 'int', 'bool' and 'checkbox' fields
	 *  @returns bool True if the object passed validation, false otherwise
	 */
	public function validate_object()
	{
		global $DEBUG;

		$this->validation_error_list = array();

		if (empty($this->obj))
			return $this->validation_error('empty object', 'Object is empty');

		if (!static::can_user_write($this->id))
			$this->validation_error('invalid object', 'The user '.get_user().' isn\'t allowed to edit this object');

		if ($this->is_template) {
			$class = $this->type . '_nacoma_Model';
			$pre = new $class($this->id);
			$pre_name = $pre->get_object_name();
			$hstprop = get_object_properties($this->db_table);
			if ($pre->obj['name'] === $hstprop['default']['template'] && $pre->obj['name'] !== $this->obj['name'])
				$this->validation_error('name', "\"{$hstprop['default']['template']}\" is a reserved {$this->type} which may not be renamed.");
		}

		$this->collapse_object(); # unset things like empty strings from web forms
		$this->expand_object(); # now, inherit everything that's unset. This sets every empty field.

		# return quickly - everything else will fail otherwise, even though this gives us
		# crappy output
		if ($this->properties['topic'] && !$this->obj[$this->properties['topic']])
			return $this->validation_error($this->properties['topic'], 'Required variable, but missing from this object');

		if ($this->properties['topic'] && $this->type !== 'service') {
			$id = (int)$this->id;
			$query = "SELECT count(id) FROM $this->db_table " .
			  "WHERE {$this->properties['topic']} = " . nacoma_db::quote_string($this->obj[$this->properties['topic']]) . " AND id != ".$id;
			$result = nacoma_db::exec_query($query);
			$row = $result->fetch();
			unset($result);
			if ($row && $row[0]) {
				$this->validation_error($this->properties['topic'], "Another object of type $this->type already exists with identifier '" . $this->obj[$this->properties['topic']] . "'");
			}

			$this->validate_object_identifier();
		}

		foreach ($this->properties['required'] as $r) {
			if ($this->obj[$r] === '') {
				// Evilness: if we have a well-defined default (such as file_id),
				// try to use it. You might feel it would look better to
				// get_default_object in the constructor so we could ignore this,
				// but we don't for two reasons:
				//   a) it's annoying to get everything to build again, because we
				//      have to construct a template object, which constructs a
				//      template object, and then you have a stack overflow.
				//   b) some code checks if (empty($obj->obj)) to check if an
				//      object was loaded, instead of the more idiomatic if ($obj->id)
				if (isset($this->properties['default'][$r])) {
					$this->normalize(array($r => $this->properties['default'][$r]));
				// if we have an alternative requirement (hostgroup_name for host_name on services),
				// then it's not as required as we thought, so we ignore it
				} else if (!isset($this->properties['req_alt'][$r]) || !isset($this->obj[$this->properties['req_alt'][$r]])) {
					$this->validation_error($r, "Required variable, but missing from this object");
				}
			}
		}

		# Ensures only one is set at a time, but *will* replace your master obj typ for you
		# This might, or might not, be what you wanted.
		foreach ($this->master_obj_candidates as $type => $key) {
			if (isset($this->obj[$key]) && $this->obj[$key] && $this->master_obj_identifier !== $key) {
				if ($this->obj[$this->master_obj_identifier])
					$this->validation_error($key, "Can't have both $key and {$this->master_obj_identifier} pointing to parents");
				else {
					$this->set_master_object_type($type);
				}
			}
		}
		$v_index = $this->properties['var_index'];

		$req_vars = $this->properties['required'];
		$spec = $this->properties['special'];

		# special magic for 'new_host' objects
		if ($this->type === 'new_host') {
			$extra_props = get_object_properties('host');
			foreach ($extra_props['var_index'] as $k => $v) {
				if (!isset($v_index[$k])) {
					$v_index[$k] = $v;
				}
			}
			foreach ($extra_props['special'] as $k => $v) {
				if (!isset($v_index[$k])) {
					continue;
				}
				if (!isset($spec[$k])) {
					$spec[$k] = $v;
				}
			}
		}

		if (!empty($this->master_obj_identifier) &&
		    empty($this->obj[$this->master_obj_identifier]))
		{
			$this->validation_error($this->master_obj_identifier, "Invalid master object link");
		}
		if(!empty($this->custom_vars))
			foreach ($this->custom_vars as $k => $v) {
				$res = $this->validate_custom_var($k, $v);
				if($res)
					$this->validation_error($k, $res);
			}

		foreach ($this->obj as $k => $v) {

			# Nagios uses ';' to denote winged comments so we mark
			# it as invalid input
			if (!is_array($v) && preg_match('/(?<!\\\\);/', $v))
				$this->validation_error($k, "Variables are not allowed to contain semi-colons, unless they're escaped with a \"\\\" character");

			# we special case the 'id' var, which is strictly
			# for internal book-keeping.
			if ($k === 'id') {
				$v_type = 'int';
			}
			else {
				if (!isset($v_index[$k])) {
					$this->validation_error($k, "Illegal variable for $this->type objects");
					continue;
				}

				$v_type = $v_index[$k];
			}

			try {
				// ucfirst = make sure it's a library
				$class = ucfirst($v_type) . '_Nacomawidget';
				// Stupid PHP! It's a static method ffs! Let me call it statically!
				$class = new $class();
				$this->obj[$k] = $class->validate($this, $k, $this->obj[$k]);
			}
			catch (widget_validation_exception $ex) {
				$this->validation_error($k, $ex->getMessage());
			}
		}

		return empty($this->validation_error_list);
	}

	/**
	 * Rudimentary validation of a custom variable
	 *
	 * @param $k variable key or name
	 * @param $v variable value
	 * @return false on success, string on error
	 */
	public function validate_custom_var($k, $v)
	{
		# hosts, services, contacts and all templates can have
		# custom variables. For timeperiods, this is handled
		# separately as custom variables are indistinguishable
		# from the flexible day/date format used in Nagios 3
		switch ($this->type) {
		 case 'host': case 'host_template':
		 case 'service': case 'service_template':
		 case 'contact': case 'contact_template':
		 case 'hostgroup': case 'management_pack':
			if($k[0] !== '_')
				return "Custom var name must start with an '_' character";
			/* fallthrough, because timeperiods have dates as keys */
		 case 'timeperiod':
			if(empty($k))
				return "Custom var with empty name not allowed";

			if(strval($v) === "")
				return "Custom var '$k' with empty value not allowed";
			return false;
		}

		return "No custom variables allowed for $this->type objects";
	}

	/**
	 * Finds all objects that can use this object type
	 */
	public function find_dependant_object_types()
	{
		$ret = array();
		$exp_objects = Module_Manifest_Model::get('nacoma_import_objects');

		foreach ($exp_objects as $obj_type => $classname) {
			$props = get_object_properties($obj_type);
			if (substr($obj_type, 0, 4) === 'new_')
			  continue;
			if (!isset($props['special']))
			  continue;

			foreach($props['special'] as $k => $v) {
				if (is_array($v))
				  continue;

				if ($v === $this->type) {
					$ret[$obj_type][] = $k;
				}
			}
		}

		return $ret;
	}

	/**
	 * Locate and return all object *types* referenced by the
	 * object type we're currently manipulating.
	 * @return Array of key => true type
	 */
	public function find_referenced_object_types()
	{
		global $Supported_Object_Types;

		$ret = array();

		$find = $this->type;

		foreach ($this->properties['var_index'] as $k => $v) {
			# object variables that lead to other objects are all
			# 'select' or 'mselect', so skip all that aren't.
			if ($v !== 'select' && $v !== 'mselect')
			  continue;

			# checkbox type mselects have predefined variables,
			# specified as arrays in the 'special' subarray of
			# the properties array. Those we cannot access here,
			# and neither do we need to, so just skip them.
			if (is_array($this->properties['special'][$k]))
			  continue;

			$ret[$this->properties['special'][$k]] = true;
		}

		return $ret;
	}

	/**
	 * Get a list of objects attached to the same master
	 * @return Array of object-id => object-name type
	 */
	public function get_fellow_slaves()
	{
		if (!$this->master_obj_id) {
			if (!$this->obj)
			  return false;
		}

		$direct = array();
		$indirect = array();
		$query = "SELECT id FROM $this->db_table WHERE $this->master_obj_identifier = ".nacoma_db::quote_string($this->master_obj_id)." AND id ".get_exclude_pending_string($this->type);
		$result = nacoma_db::exec_query($query);
		while ($row = $result->fetch(PDO::FETCH_NUM)) {
			if (static::can_user_read($row[0]))
				$direct[$row[0]] = get_object_name_by_id($this->type, $row[0]);
		}
		asort($direct);
		if (is_array($this->master_obj_candidates) && $this->master_obj_type == 'host')
		{
			// Magick.
			// This is only used for top-of-page drop-downs.
			// For objects that can be children of both hosts and host groups,
			// discover all hostgroup children that applies to host.
			$query = "SELECT $this->db_table.id FROM $this->db_table INNER JOIN hostgroup ON $this->db_table.hostgroup_name = hostgroup.id INNER JOIN host_hostgroup ON hostgroup.id = host_hostgroup.hostgroup WHERE host_hostgroup.host = $this->master_obj_id AND $this->db_table.id ".get_exclude_pending_string($this->type);
			$result = nacoma_db::exec_query($query);
			// Add magick entries to the end, so they stand out
			while($row = $result->fetch(PDO::FETCH_NUM)) {
				if (static::can_user_read($row[0]))
					$indirect[$row[0]] = get_object_name_by_id($this->type, $row[0]);
			}
			asort($indirect);
		}

		$ret = array_merge($direct, $indirect);
		return $ret;
	}

	/** Sets the id of the current object */
	public function set_object_id($id)
	{
		$this->id = $id;
		$this->obj['id'] = $id;
	}

	/**
	 * Sets the master object id of the current object
	 * @return true on success, false on errors
	 */
	public function set_master_object_id($id)
	{
		if (empty($this->master_obj_identifier))
			return false;
		if (!is_numeric($id))
			return false;

		$this->master_obj_id = $id;
		$this->obj[$this->master_obj_identifier] = $id;

		return true;
	}

	/**
	 * Loads an object from database and populates $this->obj
	 * with it.
	 * @return $this->obj, or false if object wasn't found
	 */
	public function get_object()
	{
		if (!$this->id || $this->id === 'new')
			return false;

		$query = "SELECT * FROM $this->db_table WHERE id = $this->id";
		$result = nacoma_db::exec_query($query);
		$this->obj = nacoma_db::fetch_object($result);
		unset($result);

		# We also fetch junction data for the object, if any such
		# data is available for this object type. That can be
		# determined by walking the properties table and finding
		# all variables with a display-type of 'mselect'. The sub-array
		# 'special' holds what kind of object type the mselect variable
		# links to, and the function get_junction_data() will fetch
		# the numerical indices we need to put in the object
		$vindex = $this->properties['var_index'];
		$spec = $this->properties['special'];
		foreach ($vindex as $k => $v) {
			switch ($v) {
				case 'mselect':
					if (!is_array($spec[$k])) {
						$jnc_data = get_junction_data($this->db_table, $this->id, $k, $spec[$k]);
						if ($jnc_data)
							$this->obj[$k] = $jnc_data;
					}
					break;
				case 'select':
					if (!is_array($spec[$k]) && isset($this->obj[$k]) && is_numeric($this->obj[$k])) {
						$this->obj[$k] = (int)$this->obj[$k];
					}
					break;
				case 'checkbox':
					if (isset($this->obj[$k])) {
						$this->obj[$k] = explode(',', $this->obj[$k]);
						sort($this->obj[$k]);
					}
				default:
			}
		}

		# get this objects' custom variables
		$this->custom_vars = get_custom_vars($this->db_table, $this->id);

		$this->guess_master_object_type();

		if (!empty($this->master_obj_identifier) &&
			isset($this->obj[$this->master_obj_identifier]))
		{
			$this->master_obj_id = $this->obj[$this->master_obj_identifier];
		}

		return $this->obj;
	}

	/**
	 * Writes the $this->obj to the database if it validates
	 * properly.
	 *
	 * @param $user false if this change isn't made by a user. If not, no
	 *        validation will be performed, and the change won't show up in
	 *        the changelog.
	 * @return true if everything worked out ok; false if not.
	 */
	public function save_object($user=true)
	{
		if ($this->master_obj_identifier && isset($this->obj[$this->master_obj_identifier]) && $this->obj[$this->master_obj_identifier])
			$this->set_master_object_id($this->obj[$this->master_obj_identifier]);
		else if ($this->master_obj_id)
			$this->set_master_object_id($this->master_obj_id);

		if ($user && !$this->validate_object())
			return false;

		if ($this->id === 'new' || !$this->id)
			$this->id = get_next_object_id($this->db_table);

		$this->collapse_object();
		if ($user)
			$this->update_changelog();

		// glue_objects is where the database call is made
		$rc = glue_objects(array($this->id => $this->obj), $this->db_table);

		glue_custom_vars($this->db_table, $this->id, $this->custom_vars);

		if (!$rc)
			return $this->validation_error('misc', "Glue objects failed. This is bad.");

		if (isset($this->clone_id)) {
			$class = $this->type . '_nacoma_Model';
			$orig_obj = new $class($this->clone_id);
			$dep_objs = $orig_obj->find_slave_objects();
			$dep_key = $orig_obj->find_dependant_object_types();
			foreach ($dep_objs as $type => $objs) {
				if (!is_array($objs))
					continue;
				foreach ($objs as $slave_id) {
					$class = $type.'_nacoma_Model';
					$slave_obj = new $class($slave_id);
					$cloned_slave = $slave_obj->clone_object();
					$cloned_slave->set_master_object_type($this->type);
					$cloned_slave->set_master_object_id($this->id);
					if (!empty($cloned_slave->properties['topic']))
					  $cloned_slave->obj[$cloned_slave->properties['topic']] = $slave_obj->obj[$cloned_slave->properties['topic']];
					foreach ($dep_key[$type] as $key) {
						if ($cloned_slave->obj[$key] == $orig_obj->id)
							$cloned_slave->obj[$key] = $this->id;
					}
					$cloned_slave->save_object();
					assert(empty($cloned_slave->validation_error_list));
				}
			}
		}

		return true;
	}

	/**
	 * Takes the any given object and
	 * normalizes it. That is, objects which have textual
	 * representations of objects they link to get those
	 * textual representations replaced with the linked-to
	 * objects' database id numbers in this method. Each property
	 * replaces the corresponding property in the current object.
	 *
	 * This method will only overwrite the provided keys. If you want the object
	 * to be emptied, call self::get_default_object first.
	 *
	 * @param $obj the object state to normalize, in case it's not $this->obj
	 * @param $abort_on_error If false, drop invalid variables silently.
	 * @throws NacomaRelationException
	 */
	public function normalize($obj, $abort_on_error=true)
	{
		$props = $this->properties;
		$v_index = $props['var_index'];
		$spec = $props['special'];
		$logger = op5Log::instance('nacoma');

		foreach (array_keys($obj) as $k) {
			$v = $obj[$k];
			if (isset($props['deprecated_names'][$k])) {
				unset($obj[$k]);
				$k = $props['deprecated_names'][$k];
				$obj[$k] = $v;
			}
			if (!isset($v_index[$k])) {
				if ($k[0] === '_') {
					$this->custom_vars[$k] = $v;
					continue;
				} elseif ($k === 'id') {
					continue;
				} else if ($abort_on_error) {
					throw NacomaExceptions::invalid_key($this->get_object_name(), $k, $v);
				} else {
					continue;
				}
			}
			switch($v_index[$k]) {
			 case 'select':
				if (!isset($spec[$k]) || empty($v)) {
					/* WHAT!? */
					$this->obj[$k] = $v;
				} else if (is_array($spec[$k])) {
					$v = array_search($v, $spec[$k]);
					if ($v === false) {
						if($abort_on_error) {
							throw NacomaExceptions::invalid_value($this->get_object_name(), $k, $v);
						}
					} else {
						$this->obj[$k] = $v;
					}
				} else {
					$class = $spec[$k] . '_nacoma_Model';
					$id = $class::get_id_for($v);
					if ($id) {
						$this->obj[$k] = $id;
					} else if ($k == 'file_id') {
						# no file given, and the default file doesn't exist?
						# let's just use "the first one"
						$this->obj['file_id'] = 1;
					} else if ($v && $abort_on_error) {
						throw NacomaExceptions::invalid_value($this->get_object_name(), $k, $v);
					}
				}
				break;
			case 'mselect':
				$obj[$k] = (array)$obj[$k];
				if (!isset($spec[$k])) {
					break;
				}
				$ary = array();
				if (is_array($spec[$k])) {
					foreach ($obj[$k] as $j) {
						if (isset($spec[$k][$j]))
							$ary[] = $j;
						else if ($j) {
							if ($abort_on_error)
								throw NacomaExceptions::invalid_value($this->get_object_name(), $k, $j);
							else {
								continue;
							}
						}
					}
				} elseif (is_array($obj[$k])) {
					$class = $spec[$k] . '_nacoma_Model';

					foreach ($obj[$k] as $j) {
						if (!$j)
							continue;
						$id = $class::get_id_for($j);
						if ($id) {
							$ary[$id] = $id;
						} else {
							if ($abort_on_error)
								throw NacomaExceptions::invalid_value($this->get_object_name(), $k, $j);
							else {
								$logger->log('error', sprintf("While normalizing object '%s', prop '%s' with the name (not normalized) '%s', does not resolve to any object", $this->get_object_name(), $k, $j));
								continue;
							}
						}
					}
					if (!empty($this->obj[$k])) {
						foreach ($this->obj[$k] as $i => $j) {
							if ($class::can_user_read($j)) {
								unset($this->obj[$k][$i]);
							}
						}
						$this->obj[$k] = array_merge($this->obj[$k], $ary);
					} else {
						$this->obj[$k] = $ary;
					}
					continue;
				} else if (!empty($obj[$k])) {
					if ($abort_on_error)
						throw NacomaExceptions::wrong_type($this->get_object_name(), $k, $obj[$k], $v_index[$k]);
					else
						continue;
				}
				$this->obj[$k] = $ary;
				break;
			 case 'checkbox':
				$vals = $obj[$k];
				$this->obj[$k] = array();
				if (!empty($vals)) {
					foreach ($vals as $idx => $val) {
						if (isset($spec[$k][$val]))
							$this->obj[$k][] = $val;
					}
				}
				sort($this->obj[$k]);
				break;
			 case 'bool':
				if ($obj[$k] === '')
					$this->obj[$k] = '';
				else
					$this->obj[$k] = $obj[$k]?1:0;
				break;
			 default:
				$this->obj[$k] = $v;
			}
		}
	}

	/**
	 * Takes the any given object (or the current one) and
	 * returns a denormalized version of it. That is, objects are
	 * normally stored with IDs instead of actual names to other
	 * objects. This function returns a denormalized array with all
	 * ids replaced with the foreign objects' actual names.
	 *
	 * @throws NacomaRelationException
	 * @return array
	 */
	public function denormalize($obj = false)
	{
		if ($obj === false) {
			$obj = array_merge(
				is_array($this->obj) ? $this->obj : array(),
				is_array($this->custom_vars) ? $this->custom_vars: array()
			);
		}

		if (empty($obj) || !is_array($obj))
			return array();

		$props = $this->properties;
		if (!isset($props['special']))
			return $obj;

		$spec = $props['special'];
		foreach ($props['var_index'] as $k => $type) {
			$ary = false;
			if (!isset($obj[$k]))
				continue;

			$k_spec = false;
			if (isset($props['special'][$k]))
				$k_spec = $props['special'][$k];

			switch ($type) {
			 case 'mselect':
				if (empty($obj[$k])) {
					$obj[$k] = array();
					break;
				}
				if (is_array($k_spec))
					break;
				foreach ($obj[$k] as $mlink_id) {
					$class = $k_spec . '_nacoma_Model';
					if (!$class::can_user_read($mlink_id))
						continue;
					$instance = new $class($mlink_id);
					if (!empty($instance->obj)) {
						$ary[] = $instance->get_object_name();
					# invalid many-to-many relations are really bad.
					# however, "members" and "children" are backlinks - we can only error out one way
					} elseif ($k !== 'members' && $k !== 'children' && $obj[$k]) {
						throw NacomaExceptions::invalid_value($this->get_object_name(), $k, $mlink_id);
					}
				}
				if (!empty($ary)) {
					sort($ary);
					$obj[$k] = $ary;
				}
				else
					unset($obj[$k]);
				break;
			 case 'image':
				break;
			 case 'select':
				if (!is_array($k_spec) && $obj[$k]) {
					$class = $k_spec . '_nacoma_Model';
					if (!$class::can_user_read($obj[$k])) {
						unset($obj[$k]);
						continue;
					}
					$instance = new $class($obj[$k]);
					if (empty($instance->obj))
						throw NacomaExceptions::invalid_value($this->get_object_name(), $k, $obj[$k]); # error out hard
					else
						$obj[$k] = $instance->get_object_name();
				} else {
					if($obj[$k] && !isset($k_spec[$obj[$k]])) {
						throw NacomaExceptions::invalid_value($this->get_object_name(), $k, $obj[$k]); # error out hard
					}
				}
				break;
			 case 'checkbox':
				if (empty($obj[$k])) {
					$obj[$k] = array();
					break;
				}
				sort($obj[$k]);
				break;
			case 'bool':
				if ($obj[$k] !== '')
					$obj[$k] = (bool)$obj[$k];
				break;
			case 'int':
				if (is_numeric($obj[$k]))
					$obj[$k] = (int)$obj[$k];
				else
					$obj[$k] = '';
				break;
			}
			$idx = false;
		}

		return $obj;
	}

	/**
	 * Given two data objects, remove reduntant information, making the objects
	 * suitable for being included in a changelog entry.
	 */
	public function filter_objs_for_changelog(&$pre, &$post) {
		if (!is_array($pre) || !is_array($post))
			return;

		# sometimes post objs don't contain the id, but it's always supposed to
		# be the same.
		if (isset($pre['id']))
			unset($pre['id']);
		if (isset($post['id']))
			unset($post['id']);

		if ($pre === $post) {
			$pre = array();
			$post = array();
			return;
		}
		if (empty($pre) || empty($post))
			return;

		# add nul string vars for added/deleted variables
		foreach ($pre as $k => $discard) {
			if (!isset($post[$k]))
				$post[$k] = '';
		}
		foreach ($post as $k => $discard) {
			if (!isset($pre[$k]))
				$pre[$k] = '';
		}

		# only store changes
		foreach ($pre as $oldkey => $oldval) {
			if (isset($post[$oldkey]) && $post[$oldkey] === $oldval) {
				unset($post[$oldkey]);
				unset($pre[$oldkey]);
			}
			if (isset($post[$oldkey]) && is_array($pre[$oldkey]) && is_array($post[$oldkey])) {
				$preary = array_values($pre[$oldkey]);
				$postary = array_values($post[$oldkey]);
				sort($preary);
				sort($postary);
				if ($preary == $postary) {
					unset($post[$oldkey]);
					unset($pre[$oldkey]);
				}
				else {
					foreach ($preary as $preidx => $preval) {
						if (($postidx = array_search($preval, $postary)) !== false) {
							unset($preary[$preidx]);
							unset($postary[$postidx]);
						}
					}
					$pre[$oldkey] = $preary;
					$post[$oldkey] = $postary;
				}
			}
		}
	}

	/**
	 * Used by update_changelog to do the actual insertion of entries.
	 */
	private function do_update_changelog($pre_name, $post_name, $pre, $post) {
		$user = get_user();
		$sql = 'INSERT INTO changelog(object_type, username, object_id, ' .
		       'oldname, newname';
		$values = array('type' => $this->type,
		                'name' => $user,
		                'obj' => $this->id,
		                'prename' => $pre_name,
		                'postname' => $post_name);

		if (!empty($pre)) {
			$sql .= ", oldstate";
			$values['oldstate'] = serialize($pre);
		}
		if (!empty($post)) {
			$sql .= ', newstate';
			$values['newstate'] = serialize($post);
		}
		$sql .= ") VALUES(:".implode(', :', array_keys($values)).')';

		// Workaround: nacoma_db::exec_query inserts garbage into db if exposed to
		// non-ascii characters.
		$stmt = nacoma_db::prepare_statement($sql);
		$stmt->execute($values);
	}

	/**
	 * Update the changelog with a list of changes.
	 * Saved to db in serialized format.
	 * @param $action string; Set to 'delete' if that's what you're doing
	 */
	public function update_changelog($action = false)
	{
		# This is order-sensitive, so don't muck about
		$post_name = $this->get_object_name();
		$post = $action === 'delete' ? array() : $this->obj;
		$post_custom = $action === 'delete' ? array() : $this->custom_vars;
		$pre = $this->get_object($this->id);
		$pre_custom = $this->custom_vars;
		unset($this->topic);
		$this->obj = $pre;
		$pre_name = $this->get_object_name();
		$this->obj = $post;
		unset($this->topic);
		$this->custom_vars = $post_custom;

		$v_index = $this->properties['var_index'];

		$this->filter_objs_for_changelog($pre, $post);
		$this->filter_objs_for_changelog($pre_custom, $post_custom);

		# possibly no changes, so just return early
		if ($pre === $post && $pre_custom === $post_custom)
			return;

		if (empty($pre) && empty($post) && empty($pre_custom) && empty($post_custom))
			return;

		$user = get_user();
		if (empty($user)) {
			print_r($_SERVER);
			die("You don't exist. Go away!<br />\n");
		}

		// Super-scary, super-fragile, and badly written in general. Yes.
		// When an object is saved, and users edit references to a pending
		// object, the reference needs to be saved separately from everything
		// else.
		if (!empty($pre) && !is_object_pending($this->type, $this->id)) {
			$p = $this->properties;
			foreach ($p['special'] as $name => $rel) {
				if (is_array($rel))
					continue;

				if (isset($pre[$name]) && $pre[$name]) {
					if (!is_array($pre[$name])) {
						if (is_object_pending($rel, $pre[$name])) {
							if (!isset($post[$name]))
								$post[$name] = '';
							$this->do_update_changelog($pre_name, $pre_name,
							                           array($name => $pre[$name]),
							                           array($name => $post[$name]));
							unset($pre[$name]);
							unset($post[$name]);
							continue;
						}
					}
					else {
						foreach ($pre[$name] as $idx => $val) {
							if (is_object_pending($rel, $val)) {
								$this->do_update_changelog($pre_name, $pre_name,
														   array($name => array($val)),
														   array($name => array()));
								unset($pre[$name][$idx]);
							}
						}
					}
				}
				if (isset($post[$name]) && $post[$name]) {
					if (!is_array($post[$name])) {
						if (is_object_pending($rel, $post[$name])) {
							if (!$pre[$name])
								$pre[$name] = '';
							$this->do_update_changelog($pre_name, $pre_name,
							                           array($name => $pre[$name]),
							                           array($name => $post[$name]));
							unset($pre[$name]);
							unset($post[$name]);
						}
						continue;
					}
					foreach ($post[$name] as $idx => $val) {
						if (is_object_pending($rel, $val)) {
							$this->do_update_changelog($pre_name, $pre_name,
							                           array($name => array()),
							                           array($name => array($val)));
							unset($post[$name][$idx]);
						}
					}
				}
			}
		}

		if ($pre || $post)
			$this->do_update_changelog($pre_name, $post_name, $pre, $post);
		if ($pre_custom || $post_custom)
			$this->do_update_changelog($pre_name, $post_name, $pre_custom, $post_custom);
	}

	/**
	 * Finds objects that cannot exist without this object
	 *
	 * If $type is set, returns a list of slave objects of that type.
	 * If $type isn't set, returns a map of slave types to lists of slave objects.
	 */
	public function find_slave_objects($type=false)
	{
		$ret = array();

		if (!$this->id || $this->id === 'new')
			return array();

		$slave_types = $this->slave_types;
		if (!$slave_types)
			return array();

		if ($type)
			$slave_types = in_array($type, $slave_types) ? array($type) : array();

		foreach ($slave_types as $slave) {
			$slaveclass = $slave.'_nacoma_Model';
			$slaveobj = new $slaveclass();
			$prop = $slaveobj->master_obj_candidates[$this->type];

			$query = "SELECT id FROM $slaveobj->db_table WHERE $prop = $this->id";
			$result = nacoma_db::exec_query($query);
			while ($row = $result->fetch(PDO::FETCH_NUM)) {
				$ret[$slave][] = $row[0];
			}
			unset($result);
		}

		if ($type)
			return isset($ret[$type]) ? $ret[$type] : array();
		else
			return $ret;
	}

	/**
	 * Finds all objects with a variable that uses this object
	 *
	 * @return array
	 */
	private function find_related_objects()
	{
		$rev_ref_list = get_object_reverse_reference_list($this->type);
		if(!is_array($rev_ref_list)) {
			return array();
		}

		$affected_objects = array();
		foreach($rev_ref_list as $ref_obj_type => $ref_obj_var_list) {
			$p                      = get_object_properties($ref_obj_type);
			$v_index        = $p['var_index'];
			$spec           = $p['special'];

			// set up the query data
			$multi  = false;
			$single = false;
			foreach($ref_obj_var_list as $v_name) {
			if(!isset($v_index[$v_name])) {
				bug("$v_name not set in variable index for $ref_obj_type objects<br />\n", true);
				}
				if($v_index[$v_name] === 'mselect' && !is_array($spec[$v_name])) {
					$t_var = $this->type;
					if($spec[$v_name] === $ref_obj_type &&
					 substr($v_name, -13) !== 'group_members')
						$t_var = $v_name;
					$multi[$t_var] = get_junction_table_name($ref_obj_type, $v_name);
				}
				else {
					$single[] = $v_name;
				}
			}
			$s_query        = "SELECT id FROM $ref_obj_type WHERE";
			$i		      = 0;

			if($single !== false) {
				foreach($single as $s_var) {
					if($i++ !== 0)
						$s_query .= ' OR';
					$s_query .= " $s_var = " . nacoma_db::quote_string($this->id);
				}
				$result = nacoma_db::exec_query($s_query);
				while(($row = $result->fetch(PDO::FETCH_NUM))) {
					$affected_objects[$ref_obj_type][$row[0]] = $row[0];
				}
				unset($result);
			}
			if(!empty($multi)) foreach($multi as $t_var => $jnc_table) {
				if($t_var === 'children') continue;
				$query = "SELECT * from $jnc_table " .
				  "WHERE $t_var = " . nacoma_db::quote_string($this->id);
				$result = nacoma_db::exec_query($query);
				$jnc_list = nacoma_db::fetch_result($result);
				unset($result);
				if(!empty($jnc_list)) foreach($jnc_list as $jnc_entry) {
					foreach($jnc_entry as $k => $v) {
						if($k === $t_var) continue;
						$query = "SELECT COUNT(1) from $jnc_table " .
						  "WHERE $k = $v";
						$result = nacoma_db::exec_query($query);
						$tmp = $result->fetch(PDO::FETCH_NUM);
						unset($result);
						if ($tmp[0])
							$affected_objects[$ref_obj_type][$v] = $v;
					}
				}
			}
		}
		return $affected_objects;
	}

	/**
	 * Prepare an object for drawing it.
	 * This is the place to put code generic to ALL objects.
	 * For the special cases, each object can also have their
	 * own function, but this one should always be called
	 */
	public function prepare_draw()
	{
		# if we have no id, let's assume this is a new object
		if (empty($this->obj) && (!$this->id || $this->id === 'new')) {
			$this->id = 'new';
			$this->get_default_object();
		}

		# make sure all objects that display a command link
		# also shows the corresponding '_args' variable
		foreach ($this->properties['special'] as $k => $v) {
			if ($v !== 'command')
				continue;
			if (isset($this->obj[$k]) && !isset($this->obj[$k . '_args']))
				$this->obj[$k . '_args'] = '';
		}

		$this->get_object_name();

		# Only auto-expand objects that haven't got templates
		if (!isset($this->properties['var_index']['template']) || $this->is_template || $this->id === 'new')
			$this->expand_object();

		$visible = isset($this->properties['visible'])
			? $this->properties['visible']
			: null;
		$clone_id = isset($this->clone_id) ? $this->clone_id : null;

		nacoma_object::remove_invisible_props(
			$this->id, $clone_id, $this->obj, $visible
		);
	}

	/**
	 * This is a really weird method, only called by edit.php.
	 * I'm almost certain it's doing something stupid, but simply
	 * deleting it breaks things, so perhaps there's further, compatible
	 * stupidity somewhere else?
	 *
	 * Attempt to calculate the id of the master object
	 * based on values found in the $this->obj array.
	 * @return The calculated id on success. False on failure.
	 */
	public function compute_master_obj_id()
	{
		if ($this->master_obj_id) {
			$this->set_master_object_id($this->master_obj_id);
			return $this->master_obj_id;
		}

		if (!$this->master_obj_type || !$this->obj || $this->id === 'new'
		    || empty($this->obj[$this->master_obj_identifier]))
		{
			return false;
		}

		$this->master_obj_id = $this->obj[$this->master_obj_identifier] ? $this->obj[$this->master_obj_identifier] : $this->master_obj_id;
		return $this->master_obj_id;
	}

	/**
	 * Prints debug information about the object instance.
	 * This should never be used in production.
	 */
	public function debug_print()
	{
		global $DEBUG;

		if (!$DEBUG)
		  return;

		echo "<pre>\nobject variables for $this->type object '$this->id':\n";
		print_r(get_object_vars($this));
		echo "</pre>\n";
	}

	/** Removes all template derived variables from an object */
	public function template_collapse_object()
	{
		if(!in_array($this->db_table, array('host', 'service', 'contact'))) {
			return;
		}

		if (empty($this->obj['template']) || $this->obj['template'] == $this->id) {
			return;
		}

		$tmpl_nacoma_Model = $this->db_table . '_template_nacoma_Model';
		$template = new $tmpl_nacoma_Model($this->obj['template']);
		$template->expand_object();
		if (!$template->obj) {
			return;
		}

		$this->template_obj = $template->obj;

		$v_index = $this->properties['var_index'];
		foreach ($template->obj as $k => $v) {
			if (!isset($this->obj[$k]))
			  continue;

			switch ($k) {
			 case 'file_id':
			 case 'id':
			 case 'template':
				break;
			 default:
				# note that we use '==' here instead of '===', as we want
				# '12' and 12 to evaluate to the same. We also skip
				# all _args variables as we want to save those for the
				# loop below
				if ($this->obj[$k] == $template->obj[$k] && substr($k, -5, 5) != '_args') {
					unset($this->obj[$k]);
				}
			}
		}

		# handle *command and *command_args specially,
		# since they must be treated as one unit rather
		# than as separate variables
		foreach ($this->obj as $k => $v) {
			if (substr($k, -5, 5) == "_args") {
				$cmd_var = substr($k, 0, strlen($k) - 5);

				# if check_command is set in the object we
				# just continue. If it was same as that in the
				# template it would have been removed from
				# $this->obj above.
				if (isset($this->obj[$cmd_var])) {
					continue;
				}

				# template[cmd_var] and this->obj[cmd_var] are
				# identical

				# if template doesn't have the command var, it
				# can't very well have arguments for it
				if (!isset($template->obj[$cmd_var])) {
					continue;
				}

				# template and object use identical commands
				# (which has been removed in the object), so
				# if _args are identical too, we remove them
				# from the object
				if (isset($template->obj[$k]) && ($template->obj[$k] == $this->obj[$k]) || !$this->obj[$k]) {
					unset($this->obj[$k]);
					continue;
				}

				# the only case left is that we're using same
				# check-command but different arguments, and
				# the $cmd_var was removed from $this->obj, so
				# restore it so saving doesn't cause errors
				$this->obj[$cmd_var] = $template->obj[$cmd_var];
			}
		}

		foreach ($template->custom_vars as $k => $v) {
			if (!isset($this->custom_vars[$k]))
				continue;

			if ($this->custom_vars[$k] == $v)
				unset($this->custom_vars[$k]);
		}
	}

	/**
	 * Expand an object with values from the template.
	 * This is used from the validation function, which
	 * needs to see *all* values in the object as Nagios
	 * sees them.
	 */
	public function template_expand_object($seen = array())
	{
		switch ($this->db_table) {
		 case 'host': case 'service': case 'contact':
			break;
		 default:
			return;
		}

		if (empty($this->obj['template']))
		  return;
		if (in_array($this->id, $seen, true))
			return;
		$seen[] = $this->id;

		# Fetch the template and merge the two, giving precedence
		# to the real object.
		$tmpl_nacoma_Model = $this->db_table . '_template' . '_nacoma_Model';
		$template = new $tmpl_nacoma_Model($this->obj['template']);
		if (!$template->obj)
			return;
		$template->template_expand_object($seen);

		foreach ($template->obj as $k => $v) {
			switch ($k) {
			 case 'name':
			 case 'template':
			 case 'id':
				break;
			 default:
				if (!isset($this->obj[$k]))
				  $this->obj[$k] = $v;
			}
		}

		foreach ($template->custom_vars as $k => $v) {
			if (!isset($this->custom_vars[$k]))
			  $this->custom_vars[$k] = $v;
		}
	}

	/**
	 * Collapses an object, removing variables that are
	 * identical to the nagios defaults and, optionally,
	 * to those set in the template
	 */
	public function collapse_object()
	{
		$this->template_collapse_object();
		$v_index = $this->properties['var_index'];
		$req = $this->properties['required'];
		foreach ($this->obj as $k => $v) {

			# Ignore illegal variables. validate_object()
			# will take care of them anyway
			if ($k !== 'id' && !isset($v_index[$k]))
				continue;

			if (empty($v) && $v !== '0' && $v !== 0)
			  unset($this->obj[$k]);

			if ($v === '0' || $v === 0) {
				if (isset($v_index[$k]) && $v_index[$k] === 'select' && !in_array($k, $req))
				  unset($this->obj[$k]);
			}
		}
		if (isset($v_index['register'])) {
			# I know what you're thinking. "Can I remove any of those typcasts?"
			# Well, to tell you the truth, in all this bug fixing frenzy I kind of lost track myself.
			# But being as this is nacoma, the most inbred codebase in the world,
			# and would blow your head clean off, you've got to ask yourself one question:
			#    "Do I feel lucky?"
			# Well, do ya, punk?
			$this->obj['register'] = (string)(int)!$this->is_template;
		}
	}

	/**
	 * Loads $this->obj with all variables possible to set
	 * for this object type. Variables that are set in the
	 * template used by this object get the value from the
	 * template if $template_expand is true (default)
	 * @param $template_expand bool: Use variables from template
	 */
	public function expand_object()
	{
		$this->template_expand_object();

		$vindex = $this->properties['var_index'];
		foreach ($vindex as $k => $v) {
			if (!isset($this->obj[$k])) {
				$this->obj[$k] = '';
			}
		}
	}

	/**
	 * Delete object after validating that it is safe to do so...
	 * Delete exactly one object of the type $obj_type
	 * match topic value against topic to work around pressing 'reload'
	 *
	 * This method is adapted from previous version to use object variables
	 * and NOT to print anything - only return true or array with messages to
	 * the user to be handled by the gui.
	 *
	 * This method calls banish_object() to do the actual deletion (mostly).
	 * Call banish_object() straight away if you want to force the deletion.
	 *
	 * Error or success messages can be retrieved from get_deletion_message()
	 *
	 * @return True if the object was deleted, False otherwise.
	 */
	public function delete_object()
	{
		global $DEBUG;
		$log = Op5Log::instance('auth');

		if(PHP_SAPI !== 'cli' && $DEBUG) echo "</p><h2>Requested DELETE action</h2><p>\n";

		if (!static::can_user_delete()) {
			$this->delete_message = "You're not authorized to delete {$this->type} objects";
			return false;
		}

		if (!static::can_user_write($this->id)) {
			$this->delete_message = "You're not authorized to delete {$this->get_object_name()}";
			return false;
		}

		$all = $hard = false;
		$this->get_reasons_to_not_delete($all, $hard);

		if (empty($all)) {
			return $this->banish_object();
		}

		$tmp_str = '';
		if (!empty($hard)) {
			$tmp_str .= '<p>You may not delete this '.$this->type.' for the following reasons:</p><ul>';
			foreach ($hard as $type => $ary) {
				foreach ($ary as $id => $msg) {
					$tmp_str .= "<li><a href=\"edit.php?obj_type=$type&amp;obj_id=$id\">$msg</a></li>";
				}
			}
			$tmp_str .= '</ul>';
			$this->delete_message = $tmp_str;
			return false;
		}

		if (!empty($all)) {
			$slaves = $this->find_slave_objects();
			foreach ($slaves as $type => $objs) {
				if (!isset($all[$type]))
					continue;
				foreach($objs as $id) {
					if (isset($all[$type][$id]))
						unset($all[$type][$id]);
				}
				if (empty($all[$type]))
					unset($all[$type]);
			}
			if (!empty($slaves)) {
				$tmp_str .= "<p>Deleting this {$this->type} would completely remove the following objects:</p><ul>";
				foreach ($slaves as $type => $ary) {
					foreach ($ary as $id) {
						$tmp_str .= "<li><a href=\"edit.php?obj_type=$type&amp;obj_id=$id\">$type ".get_object_name_by_id($type, $id)."</a></li>";
					}
				}
				$tmp_str .= '</ul>';
			}
			if (!empty($all)) {
				$tmp_str .= "<p>Deleting this {$this->type} would affect the following objects:</p><ul>";
				foreach ($all as $type => $ary) {
					foreach ($ary as $id => $msg) {
						$tmp_str .= "<li><a href=\"edit.php?obj_type=$type&amp;obj_id=$id\">$type ".get_object_name_by_id($type, $id)."</a></li>";
					}
				}
				$tmp_str .= '</ul>';
			}
		}

		$tmp_str .= "<br />\n" .
		  "<form action='" . $_SERVER['PHP_SELF'] . "' method='post'><div>\n" .
		  "<input type='hidden' name='action' value='force_delete' />\n" .
		  "<input type='hidden' name='obj_num' value='" . $this->id. "' />\n" .
		  "<input type='hidden' name='obj_type' value='" . $this->type . "' />\n";
		if(!empty($this->master_obj_id)) {
			$tmp_str .= "<input type='hidden' name='master_obj_id' value='" .
			  $this->master_obj_id . "' />\n";
		}
		if (isset($slaves) && !empty($slaves))
			$tmp_str .= "<input type='submit' class='button' name='x' value=\"Delete object and slaves\" title=\"Delete object and slaves\" />\n";
		else
			$tmp_str .= "<input type='submit' class='button' name='x' value=\"Delete object\" title=\"Delete object\" />\n";
		$tmp_str .= "</div></form></div>\n";

		$this->delete_message = $tmp_str;
		return false;
	}

	/**
	 * The 'real' delete function.
	 *
	 * @param $user true if this was a user-initiated delete. Set to false when
	 * doing things that shouldn't show up in the changelog, such as undoes.
	 *
	 * Error or success messages can be retrieved from get_deletion_message()
	 *
	 * @return True if the object was deleted, False otherwise.
	 */
	public function banish_object($user=true) {
		global $DEBUG;

		if ($user) {
			if (!static::can_user_delete()) {
				$this->delete_message = "You're not authorized to delete {$this->type} objects";
				return false;
			}

			if (!static::can_user_write($this->id)) {
				$this->delete_message = "You're not authorized to delete {$this->get_object_name()}";
				return false;
			}
		}

		$all = $hard = false;
		$this->get_reasons_to_not_delete($all, $hard);

		if (!empty($hard)) {
			$tmp_str = '<p>You may not delete this '.$this->type.' for the following reasons:</p><ul>';
			foreach ($hard as $type => $ary) {
				foreach ($ary as $id => $msg) {
					$tmp_str .= "<li><a href=\"edit.php?obj_type=$type&amp;obj_id=$id\">$msg</a></li>";
				}
			}
			$tmp_str .= '</ul>';
			$this->delete_message = $tmp_str;
			return false;
		}

		if ($user)
			$this->update_changelog('delete');
		$obj_del 		= array();
		$jnc_del 		= array();

		$erray			= array();

		// delete any dependent objects first, so that they can still figure
		// out what object they belonged to in changelog entries
		$slaves = $this->find_slave_objects();
		foreach ($slaves as $type => $objs) {
			$class = $type . '_nacoma_Model';
			foreach ($objs as $id) {
				$sub_obj = new $class($id);
				// this shouldn't happen
				if (!$sub_obj->banish_object($user)) {
					$this->delete_message = $sub_obj->delete_message;
					return false;
				}
				$res = $sub_obj->delete_message;
				foreach ($res['obj'] as $key => $val) {
					if (isset($obj_del[$key]))
						$obj_del[$key] += $val;
					else
						$obj_del[$key] = $val;
				}
				foreach ($res['jnc'] as $key => $val) {
					if (isset($jnc_del[$key]))
						$jnc_del[$key] += $val;
					else
						$jnc_del[$key] = $val;
				}
			}
		}

		// get reverse reference list
		$rev_ref_list = get_object_reverse_reference_list($this->type);

		if(is_array($rev_ref_list)) foreach($rev_ref_list as $ref_obj_type => $ref_obj_var_list) {
			$p 			= get_object_properties($ref_obj_type);
			$v_index 	= $p['var_index'];
			$spec 		= $p['special'];

			$affected_objects[$ref_obj_type] = false;

			// set up the query data
			$multi 	= false;
			$single = false;
			foreach($ref_obj_var_list as $v_name) {
				$jnc_table = get_junction_table_name($ref_obj_type, $v_name);
				if($v_index[$v_name] === 'mselect' && !is_array($spec[$v_name])) {
					$t_var = $this->type;
					if($spec[$v_name] === $ref_obj_type &&
					   substr($v_name, -13) !== 'group_members')
						$t_var = $v_name;
					$multi[$t_var] = $jnc_table;
				}
				else {
					$ref_nacoma_Model = $ref_obj_type . '_nacoma_Model';
					$result = nacoma_db::exec_query(
						"SELECT id FROM $ref_obj_type " .
						"WHERE $v_name = " . nacoma_db::quote_string($this->id));
					foreach ($result->fetchAll(PDO::FETCH_ASSOC) as $row) {
						$ref_obj = new $ref_nacoma_Model($row['id']);
						if (isset($ref_obj->obj['register']) && !$ref_obj->obj['register']) {
							$ref_nacoma_Model = $ref_obj_type . '_template_nacoma_Model';
							$ref_obj = new $ref_nacoma_Model($row['id']);
						}
						// variable is required, and there are no (set) alternatives
						// think templates
						if (in_array($v_name, $p['required']) && (!isset($p['req_alt'][$v_name]) || !$ref_obj->obj[$p['req_alt'][$v_name]])) {
							// this shouldn't happen
							if (!$ref_obj->banish_object($user)) {
								$this->delete_message = $ref_obj->delete_message;
								return false;
							}
							$res = $ref_obj->delete_message;
							foreach ($res['obj'] as $key => $val) {
								if (isset($obj_del[$key]))
									$obj_del[$key] += $val;
								else
									$obj_del[$key] = $val;
							}
							foreach ($res['jnc'] as $key => $val) {
								if (isset($jnc_del[$key]))
									$jnc_del[$key] += $val;
								else
									$jnc_del[$key] = $val;
							}
						}
						// variable is an alternative, and the primary requirement isn't set
						// can be triggered by dependencies
						else if (isset($p['req_alt']) && in_array($v_name, $p['req_alt']) && !$ref_obj->obj[array_search($v_name, $p['req_alt'])]) {
							// this shouldn't happen
							if (!$ref_obj->banish_object($user)) {
								$this->delete_message = $ref_obj->delete_message;
								return false;
							}
							$res = $ref_obj->delete_message;
							foreach ($res['obj'] as $key => $val) {
								if (isset($obj_del[$key]))
									$obj_del[$key] += $val;
								else
									$obj_del[$key] = $val;
							}
							foreach ($res['jnc'] as $key => $val) {
								if (isset($jnc_del[$key]))
									$jnc_del[$key] += $val;
								else
									$jnc_del[$key] = $val;
							}
						}
						else {
							unset($ref_obj->obj[$v_name]);
							$ref_obj->save_object();
							if (isset($jnc_del[$jnc_table]))
								$jnc_del[$jnc_table]++;
							else
								$jnc_del[$jnc_table] = 1;
						}
					}
					unset($result);
				}
			}

			if(!empty($multi)) {
				foreach($multi as $t_var => $jnc_table) {
					if(isset($jnc_del[$jnc_table]))
						continue;
					$query = "DELETE FROM $jnc_table " .
					  "WHERE $t_var=" . nacoma_db::quote_string($this->id);
					$result = nacoma_db::exec_query($query);
					$jnc_del[$jnc_table] = nacoma_db::affected_rows($result);
					unset($result);
				}
			}
		}

		// delete the object itself, and remove from index-table (if any)
		$result = nacoma_db::exec_query("DELETE FROM $this->db_table " .
		                         "WHERE id = ".nacoma_db::quote_string($this->id));
		$obj_del[$this->type] = nacoma_db::affected_rows($result);
		unset($result);

		if ($this->type === 'host' || $this->type === 'host_template') {
			// TODO(?): set this host's parents to be parents for those orphaned??
			// maybe later. For now we just delete it here.
			$result = nacoma_db::exec_query('DELETE FROM host_parents ' .
			                         'WHERE host=' .
			                         nacoma_db::quote_string($this->id) .
			                         ' OR parents=' .
			                         nacoma_db::quote_string($this->id));
			$jnc_del['host_parents'] = nacoma_db::affected_rows($result);
			unset($result);
		} else if ($this->type === 'hostgroup') {
			$result = nacoma_db::exec_query(
				'DELETE FROM hostgroup_members WHERE members=' .
				nacoma_db::quote_string($this->id)
			);
			$jnc_del['hostgroup_members'] = nacoma_db::affected_rows($result);
			unset($result);
		} else if ($this->type === 'servicegroup') {
			$result = nacoma_db::exec_query(
				'DELETE FROM servicegroup_members WHERE members=' .
				nacoma_db::quote_string($this->id)
			);
			$jnc_del['servicegroup_members'] = nacoma_db::affected_rows($result);
			unset($result);
		}

		$p = $this->properties;

		$v_index 	= $p['var_index'];
		$spec 		= $p['special'];

		foreach($spec as $v_name => $ref_obj_type) {
			if(isset($v_index[$v_name]) && $v_index[$v_name] === 'mselect' && !is_array($ref_obj_type)) {
				$t_var = $ref_obj_type;
				if($ref_obj_type === $this->db_table) $t_var = $v_name;
				$jnc_table = get_junction_table_name($this->db_table, $t_var);
				$query = "DELETE FROM $jnc_table " .
				  "WHERE $this->db_table=" . nacoma_db::quote_string($this->id);
				$result = nacoma_db::exec_query($query);
				if (!isset($jnc_del[$jnc_table]))
					$jnc_del[$jnc_table] = 0;
				$jnc_del[$jnc_table] += nacoma_db::affected_rows($result);
				unset($result);
			}
		}

		$this->delete_message = array('obj' => $obj_del, 'jnc' => $jnc_del);
		return true;
	}

	/**
	 * After deleting an object, calling this method will return an HTML formated
	 * representation of the changes made.
	 */
	public function get_deletion_message() {
		// if deletion failed, this will be a string
		if (!is_array($this->delete_message))
			return $this->delete_message;

		$out = array();
		# we deleted the "lowest" object first, so reverse to get the highest
		foreach(array_reverse($this->delete_message['obj'], true) as $d_obj_type => $num_objs) {
			if(empty($num_objs)) continue;
			$out[] = "<p><strong>Deleted " . $num_objs . " " . $d_obj_type
				. " object".(($num_objs===1)?'':'s')
				. ".</strong/></p>\n";
		}

		foreach(array_reverse($this->delete_message['jnc'], true) as $jnc_table => $jnc_d_rel) {
			if(!$jnc_d_rel) continue;
			$rel = explode('_', $jnc_table);
			$out[] = "<p>Deleted " . $jnc_d_rel . " " . $rel[0] . " -> " .  $rel[1]
				. (($jnc_d_rel === 1) ? ' relation' : ' relations') . "</p>\n";
		}
		$out = implode('', $out);

		if (empty($out))
			$out = "<p>Nothing deleted, but nothing went wrong. Weird.</p>";
		return $out;
	}

	/**
	 * Object deletion is sometimes dangerous or weird - this finds those cases.
	 *
	 * We separate between "hard" and "soft" reasons. A user must be allowed to
	 * override soft reasons. A user must not be allowed to override hard
	 * reasons.
	 *
	 * Hard reasons:
	 * If an object is a template that's in use, we arbitrarily decide it shouldn't be deleted.
	 * If a template object uses this object, we arbitrarily decide it shouldn't be deleted.
	 * If a hostgroup is used by a management pack, we handle it like the template case. (by hostgroup_nacoma_Model)
	 *
	 * Soft reasons:
	 * If an object is referenced by another object, we should point that out to the user.
	 *
	 * The return array has the format [type => [id => message]], so you can choose between
	 * being verbose and printing the full message, or just listing the objects that are
	 * problematic.
	 *
	 * @param $all reference Will contain a list of reasons not to delete, "soft" or "hard".
	 * @param $hard reference Will contain a list of reasons not to delete, "hard" only.
	 */
	protected function get_reasons_to_not_delete(&$all, &$hard)
	{
		$all = array();
		$hard = array();
		// object isn't saved yet, so no cow on the ice
		if (!$this->id || $this->id === 'new')
			return;

		if ($this->is_template) {
			$hstprop = get_object_properties($this->db_table);
			if ($this->obj['name'] === $hstprop['default']['template']) {
				$reason = '"' . $hstprop['default']['template'] . '"' .
					" is a reserved {$this->type} which may not be deleted.";

				$all[$this->type] = array();
				$all[$this->type][$this->id] = $reason;
				$hard[$this->type] = array();
				$hard[$this->type][$this->id] = $reason;
			}
		}

		foreach ($this->find_related_objects() as $type => $ary) {
			foreach ($ary as $id) {
				$all[$type][$id] = "$type \"".get_object_name_by_id($type, $id)."\" still references this object.";
			}
		}

		if (empty($all))
			return;

		if ($this->is_template) {
			foreach ($all as $type => $ary) {
				foreach ($ary as $id => $descr) {
					$hard[$type][$id] = $descr;
				}
			}
			return;
		}

		foreach($all as $type => $ary) {
			$tmpl_nacoma_Model = $type . '_template_nacoma_Model';
			if (!class_exists($tmpl_nacoma_Model))
				continue;
			foreach($ary as $id => $_) {
				$obj = new $tmpl_nacoma_Model($id);
				if(!empty($obj->obj))
					$all[$type . '_template'][$id] = $hard[$type . '_template'][$id] = "The template {$obj->get_object_name()} still references this object.";
			}
		}
	}

	/**
	 * Fetches ids of objects dependent on the current object for
	 * their existence.
	 * @param $sub_obj_type string; Sub-object type we're interested in
	 * (ie, 'hostdependency', 'service' etc for host objects).
	 * @param $master_link_var string; Database field linking
	 * subobject back to current object. Eg: 'host_name' from service to host.
	 * @return Array Containing ids of the subobjects or false on error
	 */
	public function get_sub_object_ids($sub_obj_type, $master_link_var)
	{
		if (!$this->id)
		  return false;

		$ret = array();
		$query = "SELECT id FROM $sub_obj_type WHERE $master_link_var = $this->id AND id ".get_exclude_pending_string($sub_obj_type);
		$result = nacoma_db::exec_query($query);
		while ($row = $result->fetch(PDO::FETCH_NUM)) {
			$ret[$row[0]] = $row[0];
		}
		unset($result);
		return $ret;
	}

	/**
	 * Determine if a user is authorized to read from a specified object.
	 */
	public static function can_user_read($id) {
		$log = Op5Log::instance('auth');
		$user = get_user();
		$user = nacoma_db::quote_string($user);

		$rules = get_nacoma_access(static::TYPE);

		//User has no access to the object type at all. Let's bail.
		if (!is_array($rules)) {
			return false;
		}

		// shortcut for most common case:
		if (in_array(NACOMA_ACCESS_FULL_ACCESS, $rules)) {
			return true;
		}
		if (!$id || $id === 'new') {
			// You can create objects? Then you should be able to read "not-yet-saved objects"
			return in_array(NACOMA_ACCESS_ADD_DELETE, $rules);
		}
		// Can user see all
		if (in_array(NACOMA_ACCESS_VIEW_ALL, $rules)) {
			return true;
		}
		// Can user view the objects he/she is contact for
		else if (in_array(NACOMA_ACCESS_VIEW_CONTACT, $rules)) {
			$objects = static::get_objects_for_user();
			if (isset($objects[$id])) {
				return true;
			} else {
				$log->log("debug", "$user cannot view " . static::TYPE . " with id " . $id);
				return false;
			}
		}
		return false;
	}

	/**
	 * Determine if a user is authorized to write to a specified object.
	 */
	public static function can_user_write($id) {
		$log = Op5Log::instance('auth');
		$user = get_user();
		$user = nacoma_db::quote_string($user);

		$rules = get_nacoma_access(static::TYPE);

		//User has no access to the object type at all. Let's bail.
		if (!is_array($rules)) {
			return false;
		}
		// shortcut for most common case:
		if (count($rules) === 1 && in_array(NACOMA_ACCESS_FULL_ACCESS, $rules))
			return true;

		//Can user create a new object? The last case is solves http_api_nacoma_Model::generate_objmap()
		if (!$id || $id === 'new' || $id < 0) {
			if (in_array(NACOMA_ACCESS_ADD_DELETE, $rules)) {;
				return true;
			}
			$log->log(
				"debug",
				"$user cannot create and delete objects of type {static::TYPE}"
			);
			return false;
		}
		//Can user edit all
		if (in_array(NACOMA_ACCESS_EDIT_ALL, $rules)) {
			return true;
		} else if (in_array(NACOMA_ACCESS_EDIT_CONTACT, $rules)) {
			//Check user have rights to edit hostgroup/servicegroup
			if(static::TYPE === 'hostgroup' || static::TYPE === 'servicegroup') {
				$type = static::TYPE.'s';
				$contacts = array();
				$host_service_groups = array();
				$contact_groups = array();
				$members = array();
				$contactgroup_members = array();
				$user_info = new contact_nacoma_Model(get_user());
				$user_contactgroups = $user_info->obj[contactgroups];
				if(static::TYPE === 'hostgroup') {
					$host_service_group = new hostgroup_nacoma_model($id);
				} else {
					$host_service_group = new servicegroup_nacoma_Model($id);
				}
				foreach ($host_service_group->obj[members] as $host_service => $host_service_id) {
					if(static::TYPE === 'hostgroup') {
						$host_service_obj = new host_nacoma_model($host_service_id);
					} else {
						$host_service_obj = new service_nacoma_Model($host_service_id);
					}
					if(isset($host_service_obj->obj[contacts])) {
						$contacts = array_merge($contacts,$host_service_obj->obj[contacts]);
					}
					if(isset($host_service_obj->obj[$type])) {
						$host_service_groups = array_merge($host_service_groups,$host_service_obj->obj[$type]);
					}
					if(isset($host_service_obj->obj[contact_groups])) {
						$contact_groups = array_merge($contact_groups,$host_service_obj->obj[contact_groups]);
					}
					foreach ($contact_groups as $contact_group => $contact_group_id) {
						$contact_group_info = new contactgroup_nacoma_Model($contact_group_id);
						if(isset($contact_group_info->obj[members])) {
							$members = array_merge($members,$contact_group_info->obj[members]);
						}
						if(isset($contact_group_info->obj[contactgroup_members])) {
							$contactgroup_members = array_merge($contactgroup_members,$contact_group_info->obj[contactgroup_members]);
						}
					}
				}
				if((in_array($id, $host_service_groups) && in_array($user_info->id, $contacts)) || (in_array($user_info->id, $members)
					|| $this->isInContact($user_contactgroups, $contactgroup_members) == true)) {
					return true;
				} else {
					return false;
				}
			} else {
				// Can user edit the objects he/she is contact for
				$objects = static::get_objects_for_user(true);
				if (isset($objects[$id])) {
					return true;
				} else {
					$log->log(
						"debug",
						"$user cannot edit {static::TYPE} with id $id"
					);
					return false;
				}
			}
		}
		return false;
	}

	/**
	 * Determine if a user is authorized to delete a specified object.
	 */
	static public function can_user_delete() {
		$rules = get_nacoma_access(static::TYPE);
		if (in_array(NACOMA_ACCESS_ADD_DELETE, $rules)) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Thin wrapper for returning user's objects of this type
	 */
	static public function get_objects_for_user($write = false) {
		throw new BadMethodCallException("Not implemented");
	}

	/**
	 * Returns a list of actions that are legal for this object type
	 */
	public function get_cmd_bar_links() {
		$cmd_links = false;
		# commands have only two variables, and it doesn't really make
		# sense to propagate either of them (since a cut 'n paste works
		# just as well), so we don't print the propagate link for them.
		if ($this->type !== 'command' && !empty($this->id) && $this->id !== 'new') {
			$cmd_links['propagate'] = "<a class='toggle_propagate' title='Select properties to propagate...'" .
				" href='#'>";
		}
		if ($this->type === 'host' || $this->type === 'service') {
			$cmd_links['dependencies'] = "<a title='Dependencies for this $this->type'" .
				" href='edit.php?obj_type=" . $this->type . "dependency&amp;master_obj_id=".html::specialchars($this->id)."'>";
			$cmd_links['escalations'] = "<a title='Notification escalations for this " . $this->type . "'" .
				" href='edit.php?obj_type=" . $this->type . "escalation&amp;master_obj_id=".html::specialchars($this->id)."'>";
		}

		# print 'advanced' or 'simple' if object has a template
		# objects that don't are always fully expanded.
		if (!empty($this->obj['template']) && $this->id != 0) {
			$cmd_links['advanced'] = "<a href='$_SERVER[PHP_SELF]?obj_type=$this->type&amp;obj_id=".html::specialchars($this->id);
			if (isset($this->master_obj_id) && $this->master_obj_id != '')
				$cmd_links['advanced'] .= "&master_obj_id=".html::specialchars($this->master_obj_id);
			$cmd_links['simple'] = $cmd_links['advanced'];

			$cmd_links['advanced'] .= "&amp;action=expand' title='Show all variables for this object, with template inheritance'>";
			$cmd_links['simple'] .= "' title='Hide template-inherited and empty variables'>";
		}

		if(!isset($_REQUEST['action']) || $_REQUEST['action'] !== 'expand')
			unset($cmd_links['simple']);
		else
			unset($cmd_links['advanced']);

		# copy is the regular "quick, slap together new object with same variables and new name"
		# clone is the more advanced "take all dependent objects and do cool stuff to them"
		# as the code has been doing a little bit of both, we're now at the stage where "copy" links to the url "clone", and "clone links to the url "copy" :(
		# create the copy link
		$cmd_links['copy'] = "<a title='Copy this " . $this->type;
		$cmd_links['copy'] .= "' href ='edit.php?action=clone&amp;obj_id=".html::specialchars($this->id)."&amp;obj_type=" . $this->type;
		if(!empty($this->master_obj_id))
			$cmd_links['copy'] .= "&amp;master_obj_id=".html::specialchars($this->master_obj_id);
		$cmd_links['copy'] .= "'>";

		$cmd_links['delete'] = "<a title='Delete this object' href='delete.php?action=delete&amp;" .
			"obj_id=".html::specialchars($this->id)."&amp;obj_type=" . $this->type;
		if(!empty($this->master_obj_id))
			$cmd_links['delete'] .= "&amp;master_obj_id=".html::specialchars($this->master_obj_id);
		$cmd_links['delete'] .= "'>";

		# new objects can't sensibly have all the links
		if ($this->id === 'new' || empty($this->id)) {
			unset($cmd_links['dependencies']);
			unset($cmd_links['escalations']);
			unset($cmd_links['copy']);
			unset($cmd_links['delete']);
		}
		if ($this->type === 'host' || $this->type === 'service') {
			$depclass = $this->type.'dependency_nacoma_Model';
			$escclass = $this->type.'escalation_nacoma_Model';
			// Give the function an integer to make it happy
			if (!$depclass::can_user_read(1)) {
				unset($cmd_links['dependencies']);
			}
			// Give the function an integer to make it happy
			if (!$escclass::can_user_read(1)) {
				unset($cmd_links['escalations']);
			}
		}
		if (!static::can_user_delete()) {
			unset($cmd_links['delete']);
		}
		if (!static::can_user_write('new')) {
			unset($cmd_links['copy']);
			unset($cmd_links['propagate']);
		}

		return $cmd_links;
	}

	/**
	 * Guess the type of the master object, if applicable,
	 * and set it.
	 */
	public function guess_master_object_type()
	{
		if (!is_array($this->master_obj_candidates))
			return;

		foreach ($this->master_obj_candidates as $type => $var) {
			if (isset($this->obj[$var]) && !empty($this->obj[$var]) && $this->obj[$var] !== '0')
				$this->set_master_object_type($type);
		}
	}

	/**
	 * Set the type of the master object for multi-parented slave objects
	 */
	public function set_master_object_type($type)
	{
		if (!isset($this->master_obj_candidates[$type]))
			bug("Trying to set invalid parent $type to $this->type object", true);

		$this->master_obj_type = $type;
		$this->master_obj_identifier = $this->master_obj_candidates[$type];
		if (isset($this->properties['visible']))
			$this->properties['visible'][] = $this->master_obj_identifier;

		// If for instance we copied a hg service to a host:
		foreach ($this->master_obj_candidates as $key) {
			if (isset($this->obj[$key]) && $key != $this->master_obj_identifier)
				unset($this->obj[$key]);
		}

		if (isset($this->obj[$this->master_obj_identifier]) && $this->obj[$this->master_obj_identifier])
			$this->master_obj_id = $this->obj[$this->master_obj_identifier];
		if ($this->master_obj_id)
			$this->obj[$this->master_obj_identifier] = $this->master_obj_id;
	}

	/**
	 * Convert an array of flags to a bitmask, depending on a flag->value map
	 *
	 * @param $bitnames array of flags
	 * @param $values map from flag to bit value
	 * @return bitmask as integer
	 */
	protected function pack_bitmask( $bitnames, $values ) {
		$val = 0;
		foreach( $bitnames as $name ) {
			if(isset($values[$name])) {
				$val += $values[$name];
			}
		}
		return $val;
	}

	/**
	 * Export with naming and values to map nagios struct names ond format
	 *
	 * Useful for Test this x operation
	 */
	public function get_nagios_struct($prefix='') {
		return array();
	}

	/**
	 * Extract custom variables from a nomralized object array
	 *
	 * Used in "test this X"
	 */
	public function extract_custom_vars($obj) {
		$vars = array();
		foreach( $obj as $k => $v ) {
			if( $k[0] == '_' ) {
				$vars[] = substr($k,1) . '=' . $v;
			}
		}
		return $vars;
	}

	/**
	 * Check user is valid contact
	 *
	 *
	 */
	public function isInContact($contactgroups, $contactgroups_members) {
		foreach ($contactgroups as $contact) {
			if (in_array($contact, $contactgroups_members)) {
				return true;
			}
		}
		return false;
	}
}
